"""Parsing and executing JavaScript code"""

import base64

from typing import Any

import regex
import js2py

# Thanks to http://regex.inginf.units.it/
arrowexp = regex.compile(r'\w[^\}]*+')


def to_ecma5_function(f: str) -> str:

    """Converts a ECMA6 function to ECMA5 format (without arrow expressions)

    :param f: ECMA6 function
    :type f: str
    :return: ECMA5 function
    :rtype: str
    """

    f = regex.sub(r'/\*.+?\*/', '', f)
    match = arrowexp.search(f)
    conv = '(function(){' + match.group(0) + '})()'
    return regex.sub(
        r'(?:s|\(s\)) => s.split\([\'"]{2}\).reverse\(\).join\([\'"]{2}\)',
        'function(s){return s.split(\'\').reverse().join(\'\')}',
        conv
    )


def atob(s: str) -> str:

    """Decodes base64 string

    :param s: Encoded data
    :type s: str
    :return: Decoded string
    :rtype: str
    """

    return base64.standard_b64decode(str(s)).decode('utf-8')


def exec_js(f: str) -> Any:

    """Executes a JavaScript function

    :param f: ECMA6 function
    :type f: str
    :return: JavaScript interpreter context
    :rtype: Any
    """

    ctx = js2py.EvalJs({'atob': atob})
    ctx.execute('window.document = { };')
    ctx.execute('window.Map = function(_i){ };')
    ctx.execute('window.setTimeout = function(_f,_t){ };')
    ctx.execute('window.setInterval = function(_f,_t){ };')
    ctx.execute('window.encodeURIComponent = function(_s){ };')
    ctx.execute(to_ecma5_function(f))
    return ctx



"""Exploring files in your server directory"""

from typing import Union, Optional, Any, List
from typing import TYPE_CHECKING

import lxml.html

from .atfile import AternosFile, FileType
if TYPE_CHECKING:
    from .atserver import AternosServer


class FileManager:

    """Aternos file manager class for viewing files structure

    :param atserv: :class:`python_aternos.atserver.AternosServer` instance
    :type atserv: python_aternos.atserver.AternosServer
    """

    def __init__(self, atserv: 'AternosServer') -> None:

        self.atserv = atserv

    def listdir(self, path: str = '') -> List[AternosFile]:

        """Requests a list of files
        in the specified directory

        :param path: Directory
        (an empty string means root), defaults to ''
        :type path: str, optional
        :return: List of :class:`python_aternos.atfile.AternosFile`
        :rtype: List[AternosFile]
        """

        path = path.lstrip('/')

        filesreq = self.atserv.atserver_request(
            f'https://aternos.org/files/{path}', 'GET'
        )
        filestree = lxml.html.fromstring(filesreq.content)

        fileslist = filestree.xpath(
            '//div[contains(concat(" ",normalize-space(@class)," ")," file ")]'
        )

        files = []
        for f in fileslist:

            ftype_raw = f.xpath('@data-type')[0]
            ftype = FileType.file \
                if ftype_raw == 'file' \
                else FileType.directory

            fsize = self.extract_size(
                f.xpath('./div[@class="filesize"]')
            )

            fullpath = f.xpath('@data-path')[0]
            filepath = fullpath[:fullpath.rfind('/')]
            filename = fullpath[fullpath.rfind('/'):]
            files.append(
                AternosFile(
                    self.atserv,
                    filepath, filename,
                    ftype, fsize
                )
            )

        return files

    def extract_size(self, fsize_raw: List[Any]) -> float:

        """Parses file size from the LXML tree

        :param fsize_raw: XPath method result
        :type fsize_raw: List[Any]
        :return: File size in bytes
        :rtype: float
        """

        if len(fsize_raw) > 0:

            fsize_text = fsize_raw[0].text.strip()
            fsize_num = fsize_text[:fsize_text.rfind(' ')]
            fsize_msr = fsize_text[fsize_text.rfind(' ') + 1:]

            try:
                return self.convert_size(float(fsize_num), fsize_msr)
            except ValueError:
                return -1.0

        return 0.0

    def convert_size(
            self,
            num: Union[int, float],
            measure: str) -> float:

        """Converts "human" file size to size in bytes

        :param num: Size
        :type num: Union[int,float]
        :param measure: Units (B, kB, MB, GB)
        :type measure: str
        :return: Size in bytes
        :rtype: float
        """

        measure_match = {
            'B': 1,
            'kB': 1000,
            'MB': 1000000,
            'GB': 1000000000
        }
        return measure_match.get(measure, -1) * num

    def get_file(self, path: str) -> Optional[AternosFile]:

        """Returns :class:`python_aternos.atfile.AternosFile`
        instance by its path

        :param path: Path to file including its filename
        :type path: str
        :return: _description_
        :rtype: Optional[AternosFile]
        """

        filepath = path[:path.rfind('/')]
        filename = path[path.rfind('/'):]

        filedir = self.listdir(filepath)
        for file in filedir:
            if file.name == filename:
                return file

        return None

    def dl_file(self, path: str) -> bytes:

        """Returns the file content in bytes (downloads it)

        :param path: Path to file including its filename
        :type path: str
        :return: File content
        :rtype: bytes
        """

        file = self.atserv.atserver_request(  # type: ignore
            'https://aternos.org/panel/ajax/files/download.php'
            'GET', params={
                'file': path.replace('/', '%2F')
            }
        )

        return file.content

    def dl_world(self, world: str = 'world') -> bytes:

        """Returns the world zip file content
        by its name (downloads it)

        :param world: Name of world, defaults to 'world'
        :type world: str, optional
        :return: Zip file content
        :rtype: bytes
        """

        resp = self.atserv.atserver_request(  # type: ignore
            'https://aternos.org/panel/ajax/worlds/download.php'
            'GET', params={
                'world': world.replace('/', '%2F')
            }
        )

        return resp.content



"""Operators, whitelist and banned players lists"""

import enum

from typing import List, Union
from typing import TYPE_CHECKING

import lxml.html

if TYPE_CHECKING:
    from .atserver import AternosServer


class Lists(enum.Enum):

    """Players list type enum"""

    whl = 'whitelist'
    whl_je = 'whitelist'
    whl_be = 'allowlist'
    ops = 'ops'
    ban = 'banned-players'
    ips = 'banned-ips'


class PlayersList:

    """Class for managing operators, whitelist and banned players lists

    :param lst: Players list type, must be
    :class:`python_aternos.atplayers.Lists` enum value
    :type lst: Union[str,Lists]
    :param atserv: :class:`python_aternos.atserver.AternosServer` instance
    :type atserv: python_aternos.atserver.AternosServer
    """

    def __init__(
            self,
            lst: Union[str, Lists],
            atserv: 'AternosServer') -> None:

        self.atserv = atserv
        self.lst = Lists(lst)

        common_whl = (self.lst == Lists.whl)
        # 1 is atserver.Edition.bedrock
        bedrock = (atserv.edition == 1)

        if common_whl and bedrock:
            self.lst = Lists.whl_be

        self.players: List[str] = []
        self.parsed = False

    def list_players(self, cache: bool = True) -> List[str]:

        """Parse a players list

        :param cache: If the function can return
        cached list (highly recommended), defaults to True
        :type cache: bool, optional
        :return: List of players nicknames
        :rtype: List[str]
        """

        if cache and self.parsed:
            return self.players

        listreq = self.atserv.atserver_request(
            f'https://aternos.org/players/{self.lst.value}',
            'GET'
        )
        listtree = lxml.html.fromstring(listreq.content)
        items = listtree.xpath(
            '//div[@class="list-item"]'
        )

        result = []
        for i in items:
            name = i.xpath('./div[@class="list-name"]')
            result.append(name[0].text.strip())

        self.players = result
        self.parsed = True
        return result

    def add(self, name: str) -> None:

        """Appends a player to the list by the nickname

        :param name: Player's nickname
        :type name: str
        """

        self.atserv.atserver_request(
            'https://aternos.org/panel/ajax/players/add.php',
            'POST', data={
                'list': self.lst.value,
                'name': name
            }, sendtoken=True
        )

        self.players.append(name)

    def remove(self, name: str) -> None:

        """Removes a player from the list by the nickname

        :param name: Player's nickname
        :type name: str
        """

        self.atserv.atserver_request(
            'https://aternos.org/panel/ajax/players/remove.php',
            'POST', data={
                'list': self.lst.value,
                'name': name
            }, sendtoken=True
        )

        for i, j in enumerate(self.players):
            if j == name:
                del self.players[i]



"""Entry point. Authorizes on Aternos
and allows to manage your account"""

import os
import re
import hashlib

from typing import List, Optional

import lxml.html

from .atserver import AternosServer
from .atconnect import AternosConnect
from .aterrors import CredentialsError


class Client:

    """Aternos API Client class object of which contains user's auth data

    :param atconn: :class:`python_aternos.atconnect.AternosConnect`
    instance with initialized Aternos session
    :type atconn: python_aternos.atconnect.AternosConnect
    """

    def __init__(
            self,
            atconn: AternosConnect,
            servers: Optional[List[str]] = None) -> None:

        self.atconn = atconn
        self.parsed = False
        self.servers: List[AternosServer] = []

        if servers:
            self.refresh_servers(servers)

    @classmethod
    def from_hashed(cls, username: str, md5: str):

        """Log in to Aternos with a username and a hashed password

        :param username: Your username
        :type username: str
        :param md5: Your password hashed with MD5
        :type md5: str
        :raises CredentialsError: If the API
        doesn't return a valid session cookie
        :return: Client instance
        :rtype: python_aternos.Client
        """

        atconn = AternosConnect()
        atconn.parse_token()
        atconn.generate_sec()

        credentials = {
            'user': username,
            'password': md5
        }

        loginreq = atconn.request_cloudflare(
            'https://aternos.org/panel/ajax/account/login.php',
            'POST', data=credentials, sendtoken=True
        )

        if 'ATERNOS_SESSION' not in loginreq.cookies:
            raise CredentialsError(
                'Check your username and password'
            )

        return cls(atconn)

    @classmethod
    def from_credentials(cls, username: str, password: str):

        """Log in to Aternos with a username and a plain password

        :param username: Your username
        :type username: str
        :param password: Your password without any encryption
        :type password: str
        :return: Client instance
        :rtype: python_aternos.Client
        """

        md5 = Client.md5encode(password)
        return cls.from_hashed(username, md5)

    @classmethod
    def from_session(
            cls,
            session: str,
            servers: Optional[List[str]] = None):

        """Log in to Aternos using a session cookie value

        :param session: Value of ATERNOS_SESSION cookie
        :type session: str
        :return: Client instance
        :rtype: python_aternos.Client
        """

        atconn = AternosConnect()
        atconn.session.cookies['ATERNOS_SESSION'] = session
        atconn.parse_token()
        atconn.generate_sec()

        return cls(atconn, servers)

    @classmethod
    def restore_session(cls, file: str = '~/.aternos'):

        """Log in to Aternos using a saved ATERNOS_SESSION cookie

        :param file: File where a session cookie
        was saved, deafults to `~/.aternos`
        :type file: str, optional
        :return: Client instance
        :rtype: python_aternos.Client
        """

        file = os.path.expanduser(file)
        with open(file, 'rt', encoding='utf-8') as f:
            saved = f.read().replace('\r\n', '\n').split('\n')

        session = saved[0].strip()

        if len(saved) > 1:
            return cls.from_session(
                session=session,
                servers=saved[1:]
            )

        return cls.from_session(session)

    @staticmethod
    def md5encode(passwd: str) -> str:

        """Encodes the given string with MD5

        :param passwd: String to encode
        :type passwd: str
        :return: Hexdigest hash of the string in lowercase
        :rtype: str
        """

        encoded = hashlib.md5(passwd.encode('utf-8'))
        return encoded.hexdigest().lower()

    def save_session(
            self,
            file: str = '~/.aternos',
            incl_servers: bool = True) -> None:

        """Saves an ATERNOS_SESSION cookie to a file

        :param file: File where a session cookie
        must be saved, defaults to `~/.aternos`
        :type file: str, optional
        :param incl_servers: If the function
        should include the servers IDs to
        reduce API requests count (recommended),
        defaults to True
        :type incl_servers: bool, optional
        """

        file = os.path.expanduser(file)
        with open(file, 'wt', encoding='utf-8') as f:

            f.write(self.atconn.atsession + '\n')
            if not incl_servers:
                return

            for s in self.servers:
                f.write(s.servid + '\n')

    def list_servers(self, cache: bool = True) -> List[AternosServer]:

        """Parses a list of your servers from Aternos website

        :param cache: If the function should use
        cached servers list (recommended), defaults to True
        :type cache: bool, optional
        :return: List of :class:`python_aternos.atserver.AternosServer` objects
        :rtype: list
        """

        if cache and self.parsed:
            return self.servers

        serverspage = self.atconn.request_cloudflare(
            'https://aternos.org/servers/', 'GET'
        )
        serverstree = lxml.html.fromstring(serverspage.content)

        servers = serverstree.xpath(
            '//div[@class="server-body"]/@data-id'
        )
        self.refresh_servers(servers)

        return self.servers

    def refresh_servers(self, ids: List[str]) -> None:

        """Replaces cached servers list creating
        :class:`AternosServer` objects by given IDs

        :param ids: Servers unique identifiers
        :type ids: List[str]
        """

        self.servers = []
        for s in ids:

            servid = s.strip()
            if servid == '':
                continue

            srv = AternosServer(servid, self.atconn)
            self.servers.append(srv)

        self.parsed = True

    def get_server(self, servid: str) -> AternosServer:

        """Creates a server object from the server ID.
        Use this instead of list_servers if you know the ID to save some time.

        :return: :class:`python_aternos.atserver.AternosServer` object
        :rtype: python_aternos.atserver.AternosServer
        """

        return AternosServer(servid, self.atconn)

    def logout(self) -> None:

        """Logouts from Aternos account"""

        self.atconn.request_cloudflare(
            'https://aternos.org/panel/ajax/account/logout.php',
            'GET', sendtoken=True
        )

    def change_username(self, value: str) -> None:

        """Changes a username in your Aternos account

        :param value: New username
        :type value: str
        """

        self.atconn.request_cloudflare(
            'https://aternos.org/panel/ajax/account/username.php',
            'POST', data={'username': value}, sendtoken=True
        )

    def change_email(self, value: str) -> None:

        """Changes an e-mail in your Aternos account

        :param value: New e-mail
        :type value: str
        :raises ValueError: If an invalid
        e-mail address was passed to the function
        """

        email = re.compile(
            r'^[A-Za-z0-9\-_+.]+@[A-Za-z0-9\-_+.]+\.[A-Za-z0-9\-]+$|^$'
        )
        if not email.match(value):
            raise ValueError('Invalid e-mail!')

        self.atconn.request_cloudflare(
            'https://aternos.org/panel/ajax/account/email.php',
            'POST', data={'email': value}, sendtoken=True
        )

    def change_password(self, old: str, new: str) -> None:

        """Changes a password in your Aternos account

        :param old: Old password
        :type old: str
        :param new: New password
        :type new: str
        """

        old = Client.md5encode(old)
        new = Client.md5encode(new)
        self.atconn.request_cloudflare(
            'https://aternos.org/panel/ajax/account/password.php',
            'POST', data={
                'oldpassword': old,
                'newpassword': new
            }, sendtoken=True
        )



"""Stores API connection session and sends requests"""

import re
import time
import random
import logging
from functools import partial

from typing import Optional, Union
from requests import Response

from cloudscraper import CloudScraper

from . import atjsparse
from .aterrors import TokenError
from .aterrors import CloudflareError
from .aterrors import AternosPermissionError

REQUA = \
    'Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 ' \
    '(KHTML, like Gecko) Chrome/99.0.4844.84 Safari/537.36 OPR/85.0.4341.47'


class AternosConnect:

    """
    Class for sending API requests bypass Cloudflare
    and parsing responses"""

    def __init__(self) -> None:

        self.session = CloudScraper()
        self.atsession = ''
        self.sec = ''
        self.token = ''

    def parse_token(self) -> str:

        """Parses Aternos ajax token that
        is needed for most requests

        :raises RuntimeWarning: If the parser
        can not find <head> tag in HTML response
        :raises CredentialsError: If the parser
        is unable to extract ajax token in HTML
        :return: Aternos ajax token
        :rtype: str
        """

        loginpage = self.request_cloudflare(
            'https://aternos.org/go/', 'GET'
        ).content

        # Using the standard string methods
        # instead of the expensive xml parsing
        head = b'<head>'
        headtag = loginpage.find(head)
        headend = loginpage.find(b'</head>', headtag + len(head))

        # Some checks
        if headtag < 0 or headend < 0:
            pagehead = loginpage
            raise RuntimeWarning(
                'Unable to find <head> tag, parsing the whole page'
            )

        # Extracting <head> content
        headtag = headtag + len(head)
        pagehead = loginpage[headtag:headend]

        try:
            text = pagehead.decode('utf-8', 'replace')
            js_code = re.findall(r'\(\(\)(.*?)\)\(\);', text)
            token_func = js_code[1] if len(js_code) > 1 else js_code[0]

            ctx = atjsparse.exec_js(token_func)
            self.token = ctx.window['AJAX_TOKEN']

        except (IndexError, TypeError) as err:
            raise TokenError(
                'Unable to parse TOKEN from the page'
            ) from err

        return self.token

    def generate_sec(self) -> str:

        """Generates Aternos SEC token which
        is also needed for most API requests

        :return: Random SEC key:value string
        :rtype: str
        """

        randkey = self.generate_aternos_rand()
        randval = self.generate_aternos_rand()
        self.sec = f'{randkey}:{randval}'
        self.session.cookies.set(
            f'ATERNOS_SEC_{randkey}', randval,
            domain='aternos.org'
        )

        return self.sec

    def generate_aternos_rand(self, randlen: int = 16) -> str:

        """Generates a random string using
        Aternos algorithm from main.js file

        :param randlen: Random string length, defaults to 16
        :type randlen: int, optional
        :return: Random string for SEC token
        :rtype: str
        """

        # a list with randlen+1 empty strings:
        # generate a string with spaces,
        # then split it by space
        rand_arr = (' ' * (randlen + 1)).split(' ')

        rand = random.random()
        rand_alphanum = self.convert_num(rand, 36) + ('0' * 17)

        return rand_alphanum[:18].join(rand_arr)[:randlen]

    def convert_num(
            self, num: Union[int, float, str],
            base: int, frombase: int = 10) -> str:

        """Converts an integer to specified base

        :param num: Integer in any base to convert.
        If it is a float started with `0,`,
        zero and comma will be removed to get int
        :type num: Union[int,float,str]
        :param base: New base
        :type base: int
        :param frombase: Given number base, defaults to 10
        :type frombase: int, optional
        :return: Number converted to a specified base
        :rtype: str
        """

        if isinstance(num, str):
            num = int(num, frombase)

        if isinstance(num, float):
            sliced = str(num)[2:]
            num = int(sliced)

        symbols = '0123456789abcdefghijklmnopqrstuvwxyz'
        basesym = symbols[:base]
        result = ''
        while num > 0:
            rem = num % base
            result = str(basesym[rem]) + result
            num //= base
        return result

    def request_cloudflare(
            self, url: str, method: str,
            params: Optional[dict] = None,
            data: Optional[dict] = None,
            headers: Optional[dict] = None,
            reqcookies: Optional[dict] = None,
            sendtoken: bool = False,
            retry: int = 5) -> Response:

        """Sends a request to Aternos API bypass Cloudflare

        :param url: Request URL
        :type url: str
        :param method: Request method, must be GET or POST
        :type method: str
        :param params: URL parameters, defaults to None
        :type params: Optional[dict], optional
        :param data: POST request data, if the method is GET,
        this dict will be combined with params, defaults to None
        :type data: Optional[dict], optional
        :param headers: Custom headers, defaults to None
        :type headers: Optional[dict], optional
        :param reqcookies: Cookies only for this request, defaults to None
        :type reqcookies: Optional[dict], optional
        :param sendtoken: If the ajax and SEC token
        should be sent, defaults to False
        :type sendtoken: bool, optional
        :param retry: How many times parser must retry
        connection to API bypass Cloudflare, defaults to 5
        :type retry: int, optional
        :raises CloudflareError:
        When the parser has exceeded retries count
        :raises NotImplementedError:
        When the specified method is not GET or POST
        :return: API response
        :rtype: requests.Response
        """

        if retry <= 0:
            raise CloudflareError('Unable to bypass Cloudflare protection')

        old_cookies = self.session.cookies
        self.session = CloudScraper()
        self.session.cookies.update(old_cookies)

        try:
            self.atsession = self.session.cookies['ATERNOS_SESSION']
        except KeyError:
            # don't rewrite atsession value
            pass

        params = params or {}
        data = data or {}
        headers = headers or {}
        reqcookies = reqcookies or {}

        method = method or 'GET'
        method = method.upper().strip()
        if method not in ('GET', 'POST'):
            raise NotImplementedError('Only GET and POST are available')

        if sendtoken:
            params['TOKEN'] = self.token
            params['SEC'] = self.sec
            headers['X-Requested-With'] = 'XMLHttpRequest'

        # requests.cookies.CookieConflictError bugfix
        reqcookies['ATERNOS_SESSION'] = self.atsession
        del self.session.cookies['ATERNOS_SESSION']

        logging.debug(f'Requesting({method}){url}')
        logging.debug(f'headers={headers}')
        logging.debug(f'params={params}')
        logging.debug(f'data={data}')
        logging.debug(f'req-cookies={reqcookies}')
        logging.debug(f'session-cookies={self.session.cookies}')

        if method == 'POST':
            sendreq = partial(
                self.session.post,
                params=params,
                data=data
            )
        else:
            sendreq = partial(
                self.session.get,
                params={**params, **data}
            )

        req = sendreq(
            url,
            headers=headers,
            cookies=reqcookies
        )

        resp_type = req.headers.get('content-type', '')
        html_type = resp_type.find('text/html') != -1
        cloudflare = req.status_code == 403

        if html_type and cloudflare:
            logging.info('Retrying to bypass Cloudflare')
            time.sleep(0.2)
            return self.request_cloudflare(
                url, method,
                params, data,
                headers, reqcookies,
                sendtoken, retry - 1
            )

        logging.debug('AternosConnect received: ' + req.text[:65])
        logging.info(
            f'{method} completed with {req.status_code} status'
        )

        if req.status_code == 402:
            raise AternosPermissionError

        req.raise_for_status()
        return req



"""Exceptions classes"""

from typing import Final


class AternosError(Exception):

    """Common error class"""


class CloudflareError(AternosError):

    """Raised when the parser is unable
    to bypass Cloudflare protection"""


class CredentialsError(AternosError):

    """Raised when a session cookie is empty
    which means incorrect credentials"""


class TokenError(AternosError):

    """Raised when the parser is unable
    to extract Aternos ajax token"""


class ServerError(AternosError):

    """Common class for server errors

    :param reason: Code which contains error reason
    :type reason: str
    :param message: Error message, defaults to ''
    :type message: str, optional
    """

    def __init__(self, reason: str, message: str = '') -> None:

        self.reason = reason
        super().__init__(message)


class ServerStartError(AternosError):

    """Raised when Aternos can not start Minecraft server

    :param reason: Code which contains error reason
    :type reason: str
    """

    MESSAGE: Final = 'Unable to start server, code: {}'
    reason_msg = {

        'eula':
            'EULA was not accepted. '
            'Use start(accepteula=True)',

        'already': 'Server is already running',
        'wrongversion': 'Incorrect software version installed',

        'file':
            'File server is unavailbale, '
            'view https://status.aternos.gmbh',

        'size': 'Available storage size limit (4 GB) was reached'
    }

    def __init__(self, reason: str) -> None:

        super().__init__(
            reason,
            self.reason_msg.get(
                reason, self.MESSAGE.format(reason)
            )
        )


class FileError(AternosError):

    """Raised when trying to execute a disallowed
    by Aternos file operation"""


# PermissionError is a built-in,
# so this exception called AternosPermissionError
class AternosPermissionError(AternosError):

    """Raised when trying to execute a disallowed command,
    usually because of shared access rights"""



"""Aternos Minecraft server"""

import enum
import json

from typing import Optional, List
from requests import Response

from .atconnect import AternosConnect
from .aterrors import ServerStartError
from .atfm import FileManager
from .atconf import AternosConfig
from .atplayers import PlayersList
from .atplayers import Lists
from .atwss import AternosWss


class Edition(enum.IntEnum):

    """Server edition type enum"""

    java = 0
    bedrock = 1


class Status(enum.IntEnum):

    """Server numeric status enum.
    It is highly recommended to use
    `AternosServer.status` instead of
    `AternosServer.status_num`"""

    off = 0
    on = 1
    starting = 2
    shutdown = 3
    unknown = 6
    error = 7
    confirm = 10


class AternosServer:

    """Class for controlling your Aternos Minecraft server

    :param servid: Unique server IDentifier
    :type servid: str
    :param atconn: :class:`python_aternos.atconnect.AternosConnect`
    instance with initialized Aternos session
    :type atconn: python_aternos.atconnect.AternosConnect
    :param reqinfo: Automatically call AternosServer.fetch()
    to get all info, defaults to `True`
    :type reqinfo: bool, optional
    """

    def __init__(
            self, servid: str,
            atconn: AternosConnect,
            reqinfo: bool = True) -> None:

        self.servid = servid
        self.atconn = atconn
        if reqinfo:
            self.fetch()

    def fetch(self) -> None:

        """Send a request to Aternos API to get all server info"""

        servreq = self.atserver_request(
            'https://aternos.org/panel/ajax/status.php',
            'GET', sendtoken=True
        )
        self._info = json.loads(servreq.content)

    def wss(self, autoconfirm: bool = False) -> AternosWss:

        """Returns :class:`python_aternos.atwss.AternosWss`
        instance for listening server streams in real-time

        :param autoconfirm: Automatically start server status listener
        when AternosWss connects to API to confirm
        server launching, defaults to `False`
        :type autoconfirm: bool, optional
        :return: :class:`python_aternos.atwss.AternosWss` object
        :rtype: python_aternos.atwss.AternosWss
        """

        return AternosWss(self, autoconfirm)

    def start(
            self,
            headstart: bool = False,
            accepteula: bool = True) -> None:

        """Starts a server

        :param headstart: Start a server in the headstart mode
        which allows you to skip all queue, defaults to `False`
        :type headstart: bool, optional
        :param accepteula: Automatically accept
        the Mojang EULA, defaults to `True`
        :type accepteula: bool, optional
        :raises ServerStartError: When Aternos
        is unable to start the server
        """

        startreq = self.atserver_request(
            'https://aternos.org/panel/ajax/start.php',
            'GET', params={'headstart': int(headstart)},
            sendtoken=True
        )
        startresult = startreq.json()

        if startresult['success']:
            return

        error = startresult['error']

        if error == 'eula' and accepteula:
            self.eula()
            self.start(accepteula=False)
            return

        raise ServerStartError(error)

    def confirm(self) -> None:

        """Confirms server launching"""

        self.atserver_request(
            'https://aternos.org/panel/ajax/confirm.php',
            'GET', sendtoken=True
        )

    def stop(self) -> None:

        """Stops the server"""

        self.atserver_request(
            'https://aternos.org/panel/ajax/stop.php',
            'GET', sendtoken=True
        )

    def cancel(self) -> None:

        """Cancels server launching"""

        self.atserver_request(
            'https://aternos.org/panel/ajax/cancel.php',
            'GET', sendtoken=True
        )

    def restart(self) -> None:

        """Restarts the server"""

        self.atserver_request(
            'https://aternos.org/panel/ajax/restart.php',
            'GET', sendtoken=True
        )

    def eula(self) -> None:

        """Accepts the Mojang EULA"""

        self.atserver_request(
            'https://aternos.org/panel/ajax/eula.php',
            'GET', sendtoken=True
        )

    def files(self) -> FileManager:

        """Returns :class:`python_aternos.atfm.FileManager`
        instance for file operations

        :return: :class:`python_aternos.atfm.FileManager` object
        :rtype: python_aternos.atfm.FileManager
        """

        return FileManager(self)

    def config(self) -> AternosConfig:

        """Returns :class:`python_aternos.atconf.AternosConfig`
        instance for editing server settings

        :return: :class:`python_aternos.atconf.AternosConfig` object
        :rtype: python_aternos.atconf.AternosConfig
        """

        return AternosConfig(self)

    def players(self, lst: Lists) -> PlayersList:

        """Returns :class:`python_aternos.atplayers.PlayersList`
        instance for managing operators, whitelist and banned players lists

        :param lst: Players list type, must be
        the :class:`python_aternos.atplayers.Lists` enum value
        :type lst: python_aternos.atplayers.Lists
        :return: :class:`python_aternos.atplayers.PlayersList`
        :rtype: python_aternos.atplayers.PlayersList
        """

        return PlayersList(lst, self)

    def atserver_request(
            self, url: str, method: str,
            params: Optional[dict] = None,
            data: Optional[dict] = None,
            headers: Optional[dict] = None,
            sendtoken: bool = False) -> Response:

        """Sends a request to Aternos API
        with server IDenitfier parameter

        :param url: Request URL
        :type url: str
        :param method: Request method, must be GET or POST
        :type method: str
        :param params: URL parameters, defaults to None
        :type params: Optional[dict], optional
        :param data: POST request data, if the method is GET,
        this dict will be combined with params, defaults to None
        :type data: Optional[dict], optional
        :param headers: Custom headers, defaults to None
        :type headers: Optional[dict], optional
        :param sendtoken: If the ajax and SEC token
        should be sent, defaults to False
        :type sendtoken: bool, optional
        :return: API response
        :rtype: requests.Response
        """

        return self.atconn.request_cloudflare(
            url=url, method=method,
            params=params, data=data,
            headers=headers,
            reqcookies={
                'ATERNOS_SERVER': self.servid
            },
            sendtoken=sendtoken
        )

    @property
    def subdomain(self) -> str:

        """Server subdomain (part of domain before `.aternos.me`)

        :return: Subdomain
        :rtype: str
        """

        atdomain = self.domain
        return atdomain[:atdomain.find('.')]

    @subdomain.setter
    def subdomain(self, value: str) -> None:

        """Set new subdomain for your server

        :param value: Subdomain
        :type value: str
        """

        self.atserver_request(
            'https://aternos.org/panel/ajax/options/subdomain.php',
            'GET', params={'subdomain': value},
            sendtoken=True
        )

    @property
    def motd(self) -> str:

        """Server message of the day,
        which is shown below its name in the servers list

        :return: MOTD
        :rtype: str
        """

        return self._info['motd']

    @motd.setter
    def motd(self, value: str) -> None:

        """Set new message of the day

        :param value: MOTD
        :type value: str
        """

        self.atserver_request(
            'https://aternos.org/panel/ajax/options/motd.php',
            'POST', data={'motd': value},
            sendtoken=True
        )

    @property
    def address(self) -> str:

        """Full server address including domain and port

        :return: Server address
        :rtype: str
        """

        return self._info['displayAddress']

    @property
    def domain(self) -> str:

        """Server domain (test.aternos.me),
        address without port number

        :return: Domain
        :rtype: str
        """

        return self._info['ip']

    @property
    def port(self) -> int:

        """Server port number

        :return: Port
        :rtype: int
        """

        return self._info['port']

    @property
    def edition(self) -> Edition:

        """Server software edition: Java or Bedrock

        :return: Software edition
        :rtype: Edition
        """

        soft_type = self._info['bedrock']
        return Edition(soft_type)

    @property
    def software(self) -> str:

        """Server software name (e.g. `Vanilla`)

        :return: Software name
        :rtype: str
        """

        return self._info['software']

    @property
    def version(self) -> str:

        """Server software version (e.g. `1.16.5`)

        :return: Software version
        :rtype: str
        """

        return self._info['version']

    @property
    def status(self) -> str:

        """Server status string (offline, loading)

        :return: Status string
        :rtype: str
        """

        return self._info['class']

    @property
    def status_num(self) -> int:

        """Server numeric status. It is highly recommended
        to use status string instead of a number.

        :return: Status code
        :rtype: Status
        """

        return Status(self._info['status'])

    @property
    def players_list(self) -> List[str]:

        """List of connected players nicknames

        :return: Connected players
        :rtype: List[str]
        """

        return self._info['playerlist']

    @property
    def players_count(self) -> int:

        """How many connected players

        :return: Connected players count
        :rtype: int
        """

        return int(self._info['players'])

    @property
    def slots(self) -> int:

        """Server slots, how many players can connect

        :return: Slots count
        :rtype: int
        """

        return int(self._info['slots'])

    @property
    def ram(self) -> int:

        """Server used RAM in MB

        :return: Used RAM
        :rtype: int
        """

        return int(self._info['ram'])



"""Connects to Aternos API websocket
for real-time information"""

import enum
import json
import asyncio
import logging

from typing import Union, Any
from typing import Tuple, Dict
from typing import Callable, Coroutine
from typing import TYPE_CHECKING

import websockets

from .atconnect import REQUA
if TYPE_CHECKING:
    from .atserver import AternosServer

OneArgT = Callable[[Any], Coroutine[Any, Any, None]]
TwoArgT = Callable[[Any, Tuple[Any, ...]], Coroutine[Any, Any, None]]
FunctionT = Union[OneArgT, TwoArgT]
ArgsTuple = Tuple[FunctionT, Tuple[Any, ...]]


class Streams(enum.Enum):

    """WebSocket streams types"""

    status = (0, None)
    queue = (1, None)
    console = (2, 'console')
    ram = (3, 'heap')
    tps = (4, 'tick')
    none = (-1, None)

    def __init__(self, num: int, stream: str) -> None:
        self.num = num
        self.stream = stream


class AternosWss:

    """Class for managing websocket connection

    :param atserv: :class:`python_aternos.atserver.AternosServer` instance
    :type atserv: python_aternos.atserver.AternosServer
    :param autoconfirm: Automatically start server status listener
    when AternosWss connects to API to confirm
    server launching, defaults to `False`
    :type autoconfirm: bool, optional
    """

    def __init__(
            self,
            atserv: 'AternosServer',
            autoconfirm: bool = False) -> None:

        self.atserv = atserv
        self.servid = atserv.servid

        cookies = atserv.atconn.session.cookies
        self.session = cookies['ATERNOS_SESSION']

        recvtype = Dict[Streams, ArgsTuple]
        self.recv: recvtype = {}
        self.autoconfirm = autoconfirm
        self.confirmed = False

        self.socket: Any
        self.keep: asyncio.Task
        self.msgs: asyncio.Task

    async def confirm(self) -> None:

        """Simple way to call AternosServer.confirm from this class"""

        self.atserv.confirm()

    def wssreceiver(
            self,
            stream: Streams,
            *args: Any) -> Callable[[FunctionT], Any]:

        """Decorator that marks your function as a stream receiver.
        When websocket receives message from the specified stream,
        it calls all listeners created with this decorator.

        :param stream: Stream that your function should listen
        :type stream: python_aternos.atwss.Streams
        :param args: Arguments which will be passed to your function
        :type args: tuple, optional
        :return: ...
        :rtype: Callable[[Callable[[Any], Coroutine[Any, Any, None]]], Any]
        """

        def decorator(func: FunctionT) -> None:
            self.recv[stream] = (func, args)
        return decorator

    async def connect(self) -> None:

        """Connect to the websocket server and start all stream listeners"""

        headers = [
            ('Host', 'aternos.org'),
            ('User-Agent', REQUA),
            (
                'Cookie',
                f'ATERNOS_SESSION={self.session}; '
                f'ATERNOS_SERVER={self.servid}'
            )
        ]
        self.socket = await websockets.connect(  # type: ignore
            'wss://aternos.org/hermes/',
            origin='https://aternos.org',
            extra_headers=headers
        )

        @self.wssreceiver(Streams.status)
        async def confirmfunc(msg):

            """Automatically confirm Minecraft server launching"""

            if not self.autoconfirm:
                return

            in_queue = (msg['class'] == 'queueing')
            pending = (msg['queue']['pending'] == 'pending')
            confirmation = in_queue and pending

            if confirmation and not self.confirmed:
                self.confirm()

        @self.wssreceiver(Streams.status)
        async def streamsfunc(msg):

            """Automatically starts streams. Detailed description:

            According to the websocket messages from the web site,
            Aternos can't receive any data from a stream (e.g. console) until
            it requests this stream via the special message
            to the websocket server: `{"stream":"console","type":"start"}`
            on which the server responses with: `{"type":"connected"}`
            Also, there are RAM (used heap) and TPS (ticks per second)
            streams that must be enabled before trying to get information.
            Enabling the stream for listening the server status is not needed,
            these data is sent from API by default, so there's None value in
            the second item of its stream type tuple
            (`<Streams.status: (0, None)>`).
            https://github.com/DarkCat09/python-aternos/issues/22#issuecomment-1146788496
            """

            if msg['status'] == 2:
                # Automatically start streams
                for strm in self.recv:

                    if not isinstance(strm, Streams):
                        continue

                    if strm.stream:
                        logging.debug(f'Enabling {strm.stream} stream')
                        await self.send({
                            'stream': strm.stream,
                            'type': 'start'
                        })

        await self.wssworker()

    async def close(self) -> None:

        """Closes websocket connection and stops all listeners"""

        self.keep.cancel()
        self.msgs.cancel()
        await self.socket.close()
        del self.socket

    async def send(self, obj: Union[Dict[str, Any], str]) -> None:

        """Sends a message to websocket server

        :param obj: Message, may be a string or a dict
        :type obj: Union[Dict[str, Any],str]
        """

        if isinstance(obj, dict):
            obj = json.dumps(obj)

        await self.socket.send(obj)

    async def wssworker(self) -> None:

        """Starts async tasks in background
        for receiving websocket messages
        and sending keepalive ping"""

        self.keep = asyncio.create_task(self.keepalive())
        self.msgs = asyncio.create_task(self.receiver())

    async def keepalive(self) -> None:

        """Each 49 seconds sends keepalive ping to websocket server"""

        try:
            while True:
                await asyncio.sleep(49)
                await self.socket.send('{"type":"\u2764"}')

        except asyncio.CancelledError:
            pass

    async def receiver(self) -> None:

        """Receives messages from websocket servers
        and calls user's streams listeners"""

        while True:
            try:
                data = await self.socket.recv()
                obj = json.loads(data)
                msgtype = Streams.none

                if obj['type'] == 'line':
                    msgtype = Streams.console
                    msg = obj['data'].strip('\r\n ')

                elif obj['type'] == 'heap':
                    msgtype = Streams.ram
                    msg = int(obj['data']['usage'])

                elif obj['type'] == 'tick':
                    msgtype = Streams.tps
                    ticks = 1000 / obj['data']['averageTickTime']
                    msg = 20 if ticks > 20 else ticks

                elif obj['type'] == 'status':
                    msgtype = Streams.status
                    msg = json.loads(obj['message'])

                if msgtype in self.recv:

                    # function info tuple
                    func: ArgsTuple = self.recv[msgtype]

                    # if arguments is not empty
                    if func[1]:
                        # call the function with args
                        coro = func[0](msg, func[1])  # type: ignore
                    else:
                        # mypy error: Too few arguments
                        # looks like bug, so it is ignored
                        coro = func[0](msg)  # type: ignore
                    # run
                    asyncio.create_task(coro)

            except asyncio.CancelledError:
                break



"""File info object used by `python_aternos.atfm`"""

import enum

from typing import Union
from typing import TYPE_CHECKING

import lxml.html

from .aterrors import FileError

if TYPE_CHECKING:
    from .atserver import AternosServer


class FileType(enum.IntEnum):

    """File or dierctory"""

    file = 0
    directory = 1


class AternosFile:

    """File class which contains info about its path, type and size

    :param atserv: :class:`python_aternos.atserver.AternosServer` instance
    :type atserv: python_aternos.atserver.AternosServer
    :param path: Path to the file
    :type path: str
    :param name: Filename
    :type name: str
    :param ftype: File or directory
    :type ftype: python_aternos.atfile.FileType
    :param size: File size, defaults to 0
    :type size: Union[int,float], optional
    """

    def __init__(
            self,
            atserv: 'AternosServer',
            path: str, name: str,
            ftype: FileType = FileType.file,
            size: Union[int, float] = 0) -> None:

        self.atserv = atserv
        self._path = path.lstrip('/')
        self._name = name
        self._full = path + name
        self._ftype = ftype
        self._size = float(size)

    def delete(self) -> None:

        """Deletes the file"""

        self.atserv.atserver_request(
            'https://aternos.org/panel/ajax/delete.php',
            'POST', data={'file': self._full},
            sendtoken=True
        )

    def get_content(self) -> bytes:

        """Requests file content in bytes (downloads it)

        :raises FileError: If downloading
        the file is not allowed by Aternos
        :return: File content
        :rtype: bytes
        """

        file = self.atserv.atserver_request(
            'https://aternos.org/panel/ajax/files/download.php',
            'GET', params={
                'file': self._full
            }
        )
        if file.content == b'{"success":false}':
            raise FileError('Unable to download the file. Try to get text')
        return file.content

    def set_content(self, value: bytes) -> None:

        """Modifies the file content

        :param value: New content
        :type value: bytes
        """

        self.atserv.atserver_request(
            'https://aternos.org/panel/ajax/save.php',
            'POST', data={
                'file': self._full,
                'content': value
            }, sendtoken=True
        )

    def get_text(self) -> str:

        """Requests editing the file as a text
        (try it if downloading is disallowed)

        :return: File text content
        :rtype: str
        """

        editor = self.atserv.atserver_request(
            f'https://aternos.org/files/{self._full.lstrip("/")}', 'GET'
        )
        edittree = lxml.html.fromstring(editor.content)

        editblock = edittree.xpath('//div[@id="editor"]')[0]
        return editblock.text_content()

    def set_text(self, value: str) -> None:

        """Modifies the file content,
        but unlike set_content takes
        a string as a new value

        :param value: New content
        :type value: str
        """

        self.set_content(value.encode('utf-8'))

    @property
    def path(self) -> str:

        """Path to a directory which
        contains the file, without leading slash

        :return: Full path to directory
        :rtype: str
        """

        return self._path

    @property
    def name(self) -> str:

        """Filename including extension

        :return: Filename
        :rtype: str
        """

        return self._name

    @property
    def full(self) -> str:

        """Absolute path to the file,
        without leading slash

        :return: Full path
        :rtype: str
        """

        return self._full

    @property
    def is_dir(self) -> bool:

        """Check if the file object is a directory

        :return: `True` if the file
        is a directory, otherwise `False`
        :rtype: bool
        """

        if self._ftype == FileType.directory:
            return True
        return False

    @property
    def is_file(self) -> bool:

        """Check if the file object is not a directory

        :return: `True` if it is a file, otherwise `False`
        :rtype: bool
        """

        if self._ftype == FileType.file:
            return True
        return False

    @property
    def size(self) -> float:

        """File size in bytes

        :return: File size
        :rtype: float
        """

        return self._size



"""
Unofficial Aternos API module written in Python.
It uses Aternos' private API and html parsing"""

from .atclient import Client
from .atserver import AternosServer
from .atserver import Edition
from .atserver import Status
from .atconnect import AternosConnect
from .atplayers import PlayersList
from .atplayers import Lists
from .atconf import AternosConfig
from .atconf import ServerOpts
from .atconf import WorldOpts
from .atconf import WorldRules
from .atconf import Gamemode
from .atconf import Difficulty
from .atwss import AternosWss
from .atwss import Streams
from .atfm import FileManager
from .atfile import AternosFile
from .atfile import FileType
from .aterrors import AternosError
from .aterrors import CloudflareError
from .aterrors import CredentialsError
from .aterrors import TokenError
from .aterrors import ServerError
from .aterrors import ServerStartError
from .aterrors import FileError
from .aterrors import AternosPermissionError
from .atjsparse import exec_js

__all__ = [

    'atclient', 'atserver', 'atconnect',
    'atplayers', 'atconf', 'atwss',
    'atfm', 'atfile',
    'aterrors', 'atjsparse',

    'Client', 'AternosServer', 'AternosConnect',
    'PlayersList', 'AternosConfig', 'AternosWss',
    'FileManager', 'AternosFile', 'AternosError',
    'CloudflareError', 'CredentialsError', 'TokenError',
    'ServerError', 'ServerStartError', 'FileError',
    'AternosPermissionError', 'exec_js',

    'Edition', 'Status', 'Lists',
    'ServerOpts', 'WorldOpts', 'WorldRules',
    'Gamemode', 'Difficulty', 'Streams', 'FileType',
]



"""Modifying server and world options"""

import enum
import re

from typing import Any, Dict, List, Union, Optional
from typing import TYPE_CHECKING

import lxml.html

if TYPE_CHECKING:
    from .atserver import AternosServer

DAT_PREFIX = 'Data:'
DAT_GR_PREFIX = 'Data:GameRules:'


class ServerOpts(enum.Enum):

    """server.options file"""

    players = 'max-players'
    gm = 'gamemode'
    difficulty = 'difficulty'
    whl = 'white-list'
    online = 'online-mode'
    pvp = 'pvp'
    cmdblock = 'enable-command-block'
    flight = 'allow-flight'
    animals = 'spawn-animals'
    monsters = 'spawn-monsters'
    villagers = 'spawn-npcs'
    nether = 'allow-nether'
    forcegm = 'force-gamemode'
    spawnlock = 'spawn-protection'
    cmds = 'allow-cheats'
    packreq = 'require-resource-pack'
    pack = 'resource-pack'


class WorldOpts(enum.Enum):

    """level.dat file"""

    seed12 = 'randomseed'
    seed = 'seed'
    hardcore = 'hardcore'
    difficulty = 'Difficulty'


class WorldRules(enum.Enum):

    """/gamerule list"""

    advs = 'announceAdvancements'
    univanger = 'universalAnger'
    cmdout = 'commandBlockOutput'
    elytra = 'disableElytraMovementCheck'
    raids = 'disableRaids'
    daynight = 'doDaylightCycle'
    entdrop = 'doEntityDrops'
    fire = 'doFireTick'
    phantoms = 'doInsomnia'
    immrespawn = 'doImmediateRespawn'
    limitcraft = 'doLimitedCrafting'
    mobloot = 'doMobLoot'
    mobs = 'doMobSpawning'
    patrols = 'doPatrolSpawning'
    blockdrop = 'doTileDrops'
    traders = 'doTraderSpawning'
    weather = 'doWeatherCycle'
    drowndmg = 'drowningDamage'
    falldmg = 'fallDamage'
    firedmg = 'fireDamage'
    snowdmg = 'freezeDamage'
    forgive = 'forgiveDeadPlayers'
    keepinv = 'keepInventory'
    deathmsg = 'showDeathMessages'
    admincmdlog = 'logAdminCommands'
    cmdlen = 'maxCommandChainLength'
    entcram = 'maxEntityCramming'
    mobgrief = 'mobGriefing'
    regen = 'naturalRegeneration'
    sleeppct = 'playersSleepingPercentage'
    rndtick = 'randomTickSpeed'
    spawnradius = 'spawnRadius'
    reducedf3 = 'reducedDebugInfo'
    spectchunkgen = 'spectatorsGenerateChunks'
    cmdfb = 'sendCommandFeedback'


class Gamemode(enum.IntEnum):

    """/gamemode numeric list"""

    survival = 0
    creative = 1
    adventure = 2
    spectator = 3


class Difficulty(enum.IntEnum):

    """/difficulty numeric list"""

    peaceful = 0
    easy = 1
    normal = 2
    hard = 3


# checking timezone format
tzcheck = re.compile(r'(^[A-Z]\w+\/[A-Z]\w+$)|^UTC$')
# options types converting
convert = {
    'config-option-number': int,
    'config-option-select': int,
    'config-option-toggle': bool
}


class AternosConfig:

    """Class for editing server settings

    :param atserv: :class:`python_aternos.atserver.AternosServer` object
    :type atserv: python_aternos.atserver.AternosServer
    """

    def __init__(self, atserv: 'AternosServer') -> None:

        self.atserv = atserv

    def get_timezone(self) -> str:

        """Parses timezone from options page

        :return: Area/Location
        :rtype: str
        """

        optreq = self.atserv.atserver_request(
            'https://aternos.org/options', 'GET'
        )
        opttree = lxml.html.fromstring(optreq)

        tzopt = opttree.xpath(
            '//div[@class="options-other-input timezone-switch"]'
        )[0]
        tztext = tzopt.xpath('.//div[@class="option current"]')[0].text
        return tztext.strip()

    def set_timezone(self, value: str) -> None:

        """Sets new timezone

        :param value: New timezone
        :type value: str
        :raises ValueError: If given string
        doesn't match Area/Location format
        """

        matches_tz = tzcheck.search(value)
        if not matches_tz:
            raise ValueError(
                'Timezone must match zoneinfo format: Area/Location'
            )

        self.atserv.atserver_request(
            'https://aternos.org/panel/ajax/timezone.php',
            'POST', data={'timezone': value},
            sendtoken=True
        )

    def get_java(self) -> int:

        """Parses Java version from options page

        :return: Java image version
        :rtype: int
        """

        optreq = self.atserv.atserver_request(
            'https://aternos.org/options', 'GET'
        )
        opttree = lxml.html.fromstring(optreq)
        imgopt = opttree.xpath(
            '//div[@class="options-other-input image-switch"]'
        )[0]
        imgver = imgopt.xpath(
            './/div[@class="option current"]/@data-value'
        )[0]

        jdkver = str(imgver or '').removeprefix('openjdk:')
        return int(jdkver)

    def set_java(self, value: int) -> None:

        """Sets new Java version

        :param value: New Java image version
        :type value: int
        """

        self.atserv.atserver_request(
            'https://aternos.org/panel/ajax/image.php',
            'POST', data={'image': f'openjdk:{value}'},
            sendtoken=True
        )

    #
    # server.properties
    #
    def set_server_prop(self, option: str, value: Any) -> None:

        """Sets server.properties option

        :param option: Option name
        :type option: str
        :param value: New value
        :type value: Any
        """

        self.__set_prop(
            '/server.properties',
            option, value
        )

    def get_server_props(self, proptyping: bool = True) -> Dict[str, Any]:

        """Parses all server.properties from options page

        :param proptyping: If the returned dict should contain value
        that matches property type (e.g. max-players will be int)
        instead of string, defaults to True
        :type proptyping: bool, optional
        :return: Server.properties dict
        :rtype: Dict[str,Any]
        """

        return self.__get_all_props('https://aternos.org/options', proptyping)

    def set_server_props(self, props: Dict[str, Any]) -> None:

        """Updates server.properties options with the given dict

        :param props: Dict with properties `{key:value}`
        :type props: Dict[str,Any]
        """

        for key in props:
            self.set_server_prop(key, props[key])

    #
    # level.dat
    #
    def set_world_prop(
            self, option: Union[WorldOpts, WorldRules],
            value: Any, gamerule: bool = False,
            world: str = 'world') -> None:

        """Sets level.dat option for specified world

        :param option: Option name
        :type option: Union[WorldOpts,WorldRules]
        :param value: New value
        :type value: Any
        :param gamerule: If the option
        is a gamerule, defaults to False
        :type gamerule: bool, optional
        :param world: Name of the world which
        level.dat must be edited, defaults to 'world'
        :type world: str, optional
        """

        prefix = DAT_PREFIX
        if gamerule:
            prefix = DAT_GR_PREFIX

        self.__set_prop(
            f'/{world}/level.dat',
            f'{prefix}{option}',
            value
        )

    def get_world_props(
            self, world: str = 'world',
            proptyping: bool = True) -> Dict[str, Any]:

        """Parses level.dat from specified world's options page

        :param world: Name of the world, defaults to 'world'
        :type world: str, optional
        :param proptyping: If the returned dict should contain the value
        that matches property type (e.g. randomTickSpeed will be bool)
        instead of string, defaults to True
        :type proptyping: bool, optional
        :return: Level.dat dict
        :rtype: Dict[str,Any]
        """

        return self.__get_all_props(
            f'https://aternos.org/files/{world}/level.dat',
            proptyping, [DAT_PREFIX, DAT_GR_PREFIX]
        )

    def set_world_props(
            self,
            props: Dict[Union[WorldOpts, WorldRules], Any],
            world: str = 'world') -> None:

        """Sets level.dat options from
        the dictionary for the specified world

        :param props: Level.dat options
        :type props: Dict[Union[WorldOpts, WorldRules], Any]
        :param world: name of the world which
        level.dat must be edited, defaults to 'world'
        :type world: str
        """

        for key in props:
            self.set_world_prop(
                option=key,
                value=props[key],
                world=world
            )

    #
    # helpers
    #
    def __set_prop(self, file: str, option: str, value: Any) -> None:

        self.atserv.atserver_request(
            'https://aternos.org/panel/ajax/config.php',
            'POST', data={
                'file': file,
                'option': option,
                'value': value
            }, sendtoken=True
        )

    def __get_all_props(
            self, url: str, proptyping: bool = True,
            prefixes: Optional[List[str]] = None) -> Dict[str, Any]:

        optreq = self.atserv.atserver_request(url, 'GET')
        opttree = lxml.html.fromstring(optreq.content)
        configs = opttree.xpath('//div[@class="config-options"]')

        for i, conf in enumerate(configs):
            opts = conf.xpath('/div[contains(@class,"config-option ")]')
            result = {}

            for opt in opts:
                key = opt.xpath(
                    './/span[@class="config-option-output-key"]'
                )[0].text
                value = opt.xpath(
                    './/span[@class="config-option-output-value"]'
                )[0].text

                if prefixes is not None:
                    key = f'{prefixes[i]}{key}'

                opttype = opt.xpath('/@class').split(' ')[1]
                if proptyping and opttype in convert:
                    value = convert[opttype](value)

                result[key] = value

        return result



import setuptools

with open('README.md', 'rt') as readme:
    long_description = readme.read()

setuptools.setup(
    name='python-aternos',
    version='1.1.2',
    author='Chechkenev Andrey (@DarkCat09)',
    author_email='aacd0709@mail.ru',
    description='An unofficial Aternos API',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/DarkCat09/python-aternos',
    project_urls={
        'Homepage': 'https://codeberg.org/DarkCat09/python-aternos',
        'GitHub': 'https://github.com/DarkCat09/python-aternos',
        'Bug Tracker': 'https://github.com/DarkCat09/python-aternos/issues',
        'Bug Tracker 2': 'https://codeberg.org/DarkCat09/python-aternos/issues'
    },
    classifiers=[
        'Development Status :: 4 - Beta',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: OS Independent',
        'Operating System :: Microsoft :: Windows',
        'Operating System :: POSIX :: Linux',
        'Operating System :: MacOS',
        'Intended Audience :: Developers',
        'Topic :: Internet',
        'Typing :: Typed'
    ],
    install_requires=[
        'lxml>=4.8.0',
        'cloudscraper>=1.2.60',
        'js2py>=0.71',
        'websockets>=10.1',
        'regex>=2022.3.15'
    ],
    packages=['python_aternos'],
    python_requires=">=3.7",
)
